/***********************************************************************
 * 7 szegmenses kijelző vezérlése RP2040 segítségével
 * 
 * Platform:    Arduino MbedOS RP2040 Boards
 * Leírás:       
 *      Ez a program egy 5463AS típusú (közös katódú)
 *      4 digites multiplex kijelző kezelésére
 *      szolgál az RP2040 PIO (Programmable Input/Output) 
 *      perifériája és hardveres logikája segítségével. 
 *
 * Fő funkciók:
 *    - Számok kódolása és megjelenítése 7 szegmenses formátumban
 *    - Multiplex vezérlés PIO és side-set logika segítségével
 *    - FIFO-n keresztüli adatmozgatás a PIO és memória között
 *
 * Hardver feltételek:
 *    - Raspberry Pi Pico (RP2040) mikrovezérlő
 *    - 4 digites 7 szegmenses kijelző (5463AS)
 *    - Ellenállások a szegmensek áramkorlátozásához
 *
 * Felhasznált GPIO lábak:
 *  - GPIO 0–7: Szegmensek vezérlése (A–G, DP)
 *  - GPIO 8–11: Digitek vezérlése (D1–D4)
 *
 * Bekötési lista:
 *  RP2040       5463AS kijelző
 * ==================================   
 *  - GPIO 0:  pin 11 Szegmens A
 *  - GPIO 1:  pin  7 Szegmens B
 *  - GPIO 2:  pin  4 Szegmens C
 *  - GPIO 3:  pin  2 Szegmens D
 *  - GPIO 4:  pin  1 Szegmens E
 *  - GPIO 5:  pin 10 Szegmens F
 *  - GPIO 6:  pin  5 Szegmens G
 *  - GPIO 7:  pin  3 Tizedespont (DP)
 *  - GPIO 8:  pin  6 Digit 1 aktiválása
 *  - GPIO 9:  pin  8 Digit 2 aktiválása
 *  - GPIO 10: pin  9 Digit 3 aktiválása
 *  - GPIO 11: pin 12 Digit 4 aktiválása
 *
 ***********************************************************************/

#include "hardware/pio.h"
#include "pio_7seg.pio.h" // Generált PIO fejlécfájl

// Konstansok
const uint base_pin = 0; // GPIO 0 az első szegmensvezérlő
const uint sm = 0;       // PIO állapotgép index

// Globális változók
PIO pio = pio0;          // Az első PIO példány használata

// Számok kódolása szegmensekre (a G, F, E, D, C, B, A sorrendben)
const uint8_t digit_segments[10] = {
    0b00111111, // 0
    0b00000110, // 1
    0b01011011, // 2
    0b01001111, // 3
    0b01100110, // 4
    0b01101101, // 5
    0b01111101, // 6
    0b00000111, // 7
    0b01111111, // 8
    0b01101111  // 9
};

uint counter = 0; // Kezdő érték

// Szám megjelenítése a kijelzőn
void display_number(uint number) {
    uint32_t data = 0;
    // 4 digit kódolása (legkisebb helyiérték az első)
    for (int i = 0; i < 4; i++) {
        uint digit = number % 10;       // Szám kiszámítása
        number /= 10;                   // Következő számra lépünk
        data |= (digit_segments[digit] << (8 * i)); // Bájtok összeállítása
    }
    // Adat küldése a PIO FIFO-ba
    pio_sm_put_blocking(pio, sm, data); // SDK függvény hívása
}

void setup() {
    uint offset = pio_add_program(pio, &pio_7seg_program);
    pio_7seg_init(pio, sm, offset, base_pin);
}


void loop() {
    display_number(counter); // Szám megjelenítése
    counter = (counter + 1) % 10000;    // Szám növelése (max 9999)
    delay(100);
   
}
