/*
 * Project: PIO-Based DHT Sensor Reader
 * Adapted from Valentin Milea's original example code
 * 
 * Description:
 * This program reads temperature and humidity data from a DHT sensor 
 * (DHT11, DHT12, DHT21, or DHT22) using a PIO-based implementation for 
 * Raspberry Pi Pico in the Arduino MbedOS RP2040 Boards environment.
 *
 * The program outputs the temperature in Celsius and the relative humidity 
 * as a percentage to the Serial Monitor every 2 seconds.
 *
 * Connections:
 * - Data Pin: GP8 (changeable via DATA_PIN constant)
 * - Power: 3.3V (VDD)
 * - Ground: GND
 *
 * Notes:
 * - Uses PIO and DMA for efficient sensor communication.
 * - Requires pull-up resistor for the data line (internal pull-up enabled by default).
 *
 * Copyright (c) 2021 Valentin Milea <valentin.milea@gmail.com>
 * SPDX-License-Identifier: MIT
 * 
 * Tested with Arduino IDE 1.8.19 and Arduino MbedOS RP2040 Boars v4.2.1
 */


#include <Arduino.h>
#include <pio_dht.h>

// change this to match your setup
static const dht_model_t DHT_MODEL = DHT22;
static const uint DATA_PIN = 8;

// Globális DHT struktúra
dht_t dht;

void setup() {
  Serial.begin(115200); // Soros monitor inicializálása
  delay(1000);          // Várakozás az inicializálásra
  Serial.println("\nDHT test");

  // DHT inicializálása
  dht_init(&dht, DHT_MODEL, pio0, DATA_PIN, true /* pull_up */);
}

void loop() {
  dht_start_measurement(&dht);
  float humidity;
  float temperature_c;
  dht_result_t result = dht_finish_measurement_blocking(&dht, &humidity, &temperature_c);

  if (result == DHT_RESULT_OK) {
    Serial.print(temperature_c, 1); // Celsius érték 1 tizedes pontossággal
    Serial.print(" C, ");
    Serial.print(humidity, 1); // Páratartalom 1 tizedes pontossággal
    Serial.println("% humidity");
  } else if (result == DHT_RESULT_TIMEOUT) {
    Serial.println("DHT sensor not responding. Please check your wiring.");
  } else {
    assert(result == DHT_RESULT_BAD_CHECKSUM);
    Serial.println("Bad checksum");
  }
  delay(2000); // 2 másodperces mintavétel
}
